
import os, zipfile, glob
import subprocess
from os.path import isabs, join, relpath
from contextlib import contextmanager

from .globs import translate_pattern


def shell(cmd, check=True, print_output=True):
    """
    Execute cmd.  Raises an exception if the command does not return 0.
    """
    proc = subprocess.run(cmd, shell=True, check=False, stderr=subprocess.STDOUT,
                          stdout=subprocess.PIPE)
    output = proc.stdout.decode('utf-8').strip()
    if print_output and output:
        print(output)
    if check:
        proc.check_returncode()
    return (proc.returncode, output)


def ensurelist(paramname, val, comment='#'):
    """
    Returns `val` as a list of strings.

    If `val` is already a list of strings, it is returned as-is.  If a string
    is passed, it is split into lines and blink and (optionally) comment lines
    are removed.

    paramname
      The parameter name to put into error messages.

    comment
      The single-line comment identifier.  Any lines starting with this are
      removed.  Note that end-of-line comments are *not* removed.
    """

    if isinstance(val, list):
        for line in val:
            if not isinstance(line, str):
                typename = type(line).__qualname__
                raise TypeError(f"{paramname} should be a list of strings.  Found: {typename}")
        return val

    if isinstance(val, str):
        lines = (line.strip() for line in val.strip().split('\n'))
        lines = (line for line in lines if line)
        if comment:
            lines = (line for line in lines if not line.startswith())
        return lines

    raise TypeError(f'{paramname} must be a list of strings or a string')


def archive(filename, *, patterns, root=None, nocompression=['*.png', '*.jpg'],
            exclude_dirs=['.git']):
    """
    Creates a zip file.

    * root: An optional directory used as the parent of all relative paths in
      other parameters.  If not provided, the current working directory is used.

    * patterns: A list of glob patterns to be included and excluded.

      For convenience, a triple quoted string is also accepted.  If used, each
      glob must be on its own line.  Blank lines and lines that begin with '#'
      (comments) are ignored.

      By default, a glob describes files to be added.  If a glob begins with '!'
      (followed by optional whitespace), it defines files to be excluded.  Note
      that exclusions take precedence over inclusions.  This is by design since
      it is usually convenient to use broad inclusions and specific exclusions.

    * exclude_dirs: Directory names, not paths or globs, that should be excluded.
      Excluding directories like .git and node_modules this way is significantly
      faster than excluding using a negated pattern.

    If a glob pattern starts with "/", it is considered an absolute pattern and
    is matched from the root directory.  Otherwise it is a relative pattern and
    will match anywhere under the root directory.

    For example, "/test.log" is absolute and will only match the file "test.log"
    in the root directory.  It can include subdirectoreis also like
    "/tests/*.py".  Otherwise, it can match anywhere in the path below the root.
    So "test.log" will match "<root>/test.log" and "<root>/subdir/test.log".
    Likewise, "tests/*.py" (with no leading slash) will match
    "<root>/tests/t.py" and "<root>/subdir/tests/t.py".

    Patterns are not recursive by default.  Use '**' to introduce recursion,
    similar to .gitignore and similar tools.  So "/tests/**" will match *all*
    files in "<root>/tests/" and all subdirectories.  The pattern "tests/**"
    will match all files under any "tests" directory.  "a/**/b/*.log" will match
    any log in any directory named "b" that is somewhere under an a, such as
    "a/b/test.log" and "a/x/y/z/b/test.log".
    """
    if not root:
        root = os.getcwd()

    if isinstance(patterns, str):
        lines = (line.strip() for line in patterns.split('\n'))
        nonblank = (line for line in lines
                    if line and not line.startswith('#'))
        patterns = list(nonblank)

    arc = zipfile.ZipFile(filename, 'w', compression=zipfile.ZIP_DEFLATED)

    includes = [p for p in patterns if p and not p.startswith('!')]
    excludes = [p for p in patterns if p and p.startswith('!')]

    include_regexps = [translate_pattern(p) for p in includes]
    exclude_regexps = [translate_pattern(p[1:].lstrip()) for p in excludes]
    nocomp_regexps  = [translate_pattern(p) for p in (nocompression or [])]

    exclude_dirs = set(exclude_dirs or [])

    def _matches_pattern(fn, regexps):
        for regexp in regexps:
            if regexp.search(fn):
                return True
        return False

    for parent, dirs, files in os.walk(root):
        if excludes and dirs:
            # See if we can exclude any of the directories to speed things up.
            remove = [dir for dir in dirs
                      if dir in exclude_dirs or _matches_pattern(join(parent, dir), exclude_regexps)
            ]
            for dir in remove:
                dirs.remove(dir)

        for fn in files:
            fqn = join(parent, fn)
            rn  = relpath(fqn, root)

            if not _matches_pattern(rn, include_regexps) or _matches_pattern(rn, exclude_regexps):
                continue

            compression = (zipfile.ZIP_STORED if not _matches_pattern(rn, nocomp_regexps) else None)
            arc.write(fqn, rn, compress_type=compression)


    #  for spec in includes:
    #      if isabs(spec):
    #          raise ValueError(
    #              f'Include specifications cannot be absolute.  ({spec!r})'
    #          )
    #      spec = join(root, spec)
    #      recursive = not spec.endswith('/*')

    #      for fn in glob.iglob(spec, recursive=recursive):

    #          if _matches_pattern(fn, exclude_regexps):
    #              continue

    #          compression = zipfile.ZIP_STORED
    #          if _matches_pattern(fn, nocomp_regexps):
    #              compression = None

    #          arcname = relpath(fn, root)
    #          arc.write(fn, arcname, compress_type=compression)

    return filename


@contextmanager
def chdir(dir):
    """
    A context manager that changes the directory to `dir`, then restores the previous
    directory when exited.
    """
    prev = os.getcwd()
    try:
        os.chdir(dir)
        yield
    finally:
        os.chdir(prev)




import os, logging, tempfile
from os.path import basename, exists, join
from collections import namedtuple, deque

from .clientcnxn import ClientConnection
from .protocol import hash_file

__all__ = ['Client', 'File', 'Template']

logger = logging.getLogger('pepperssh')

Stat = namedtuple('Stat', 'exists isdir isfile size hash fqn')


class Client:
    """
    Creates an SSH connection to a remote host and allows Python scripts to be
    sent and executed.

    To run a script remotely, call `script(filename)`.  This will execute the
    module and return a proxy to it.  The proxy allows you to call functions in
    the script remotely:

        script = client.script('setupdb.py')
        result = script.createdb(name='test')

    Parameters sent to the proxied function and its return value are pickled,
    so many Python objects can be used.  Since all exception classes may not be
    available on the local machine, exceptions are returned as string
    tracebacks and a RuntimeError is raised.

    Anything printed to the screen by the remote function is printed on the
    screen locally.

    Bi-directional input is not supported, so remote functions must not call
    functions like getpass().

    By default, the client stops waiting for remote function calls after 2
    minutes and disconnects.  The current command will stay running (it is not
    forcefully terminated) but will exit when the command is complete.  To
    override this default, use the function proxy's `settimeout` method:

        script.createdb.timeout = 5 * 60
        script.createdb('test')
    """
    def __init__(self, host):
        """
        Begins the connection to the remote host and starts up a server on it
        to accept commands from this client.
        """
        self.host = host
        self.cnxn = ClientConnection(host)
        self.cnxn.connect()

    def script(self, filename, timeout=10.0):
        """
        Sends the file to the remote host and returns a proxy to it.

        The file is loaded and executed on the remote host from a temporary
        working directory.  Any output printed to the screen is printed on the
        screen locally.
        """
        proxy = FileProxy(self, filename)

        sftp = self.cnxn.sshclient.open_sftp()
        rname = join(self.cnxn.rpath, basename(filename))
        logger.debug('copying %s --> %s', filename, rname)
        sftp.put(filename, rname)

        self.cnxn.send_command({
            'msgtype': 'script',
            'modulename': proxy._modulename,
            'filename': rname
        }, timeout=timeout)

        return proxy

    def _call(self, modulename, funcname, args, kwargs, *, timeout):
        # This isn't terribly efficient, but I'm trying to get something
        # working right now.  Convert args from a tuple to a list so we can
        # replace entries.
        assert isinstance(args, tuple)
        args = list(args)

        tmplcache = set()
        # The _copy_call_files function will *replace* all File and Template instances with
        # remote filenames after copying.  If the caller didn't keep another copy of the
        # Templates, they will be freed and will delete the temporary file it created with the
        # interpolated results.  We'll hang on to them here until the call completes.

        self._copy_call_files(args, kwargs, tmplcache)

        return self.cnxn.send_command({
            'msgtype': 'call',
            'mname': modulename,
            'fname': funcname,
            'args': args,
            'kwargs': kwargs
        }, timeout=timeout)

    def _get_cache_dir(self, sftp):
        DIR = '.peppercache'

        stat = self._stat(DIR)
        if not stat.exists:
            sftp.mkdir(DIR)
            stat = self._stat(DIR)

        return stat.fqn

    def _stat(self, filename):
        """
        Returns statistics about a remote file or directory, siliar to os.stat.
        """
        msg = self.cnxn.send_command({
            'msgtype': 'stat',
            'filename': filename
        }, timeout=5)
        return Stat(*(msg.get(field) for field in Stat._fields))

    def _copy_call_files(self, args, kwargs, tmplcache):
        """
        If there are any File instances in the arguments, transfer the file to
        the working directory and replace the argument with the remote
        filename.
        """
        sftp = None
        remotedir = None

        # This is wasteful if there are no File objects, but we don't know the remote name
        # until we call open_sftp().  Can we get it without doing this?

        map_local_to_remote = {}
        # Using a dictionary in case a filename is included more than once.

        queue = deque()
        queue.append(args)
        queue.append(kwargs)

        while queue:
            arg = queue.popleft()
            assert isinstance(arg, (list, dict))

            if isinstance(arg, list):
                iterator = enumerate(arg)
            else:
                iterator = arg.items()

            for key, val in iterator:
                if isinstance(val, File):
                    if not sftp:
                        sftp = self.cnxn.sshclient.open_sftp()
                        remotedir = self._get_cache_dir(sftp)

                    if isinstance(val, Template):
                        tmplcache.add(val)

                    rname = join(remotedir, basename(val.filename))
                    map_local_to_remote[val.filename] = rname
                    arg[key] = rname
                elif isinstance(val, (list, dict)):
                    queue.append(val)

        self._transfer_files(sftp, map_local_to_remote)

    def _transfer_files(self, sftp, files):
        # Paramiko is an absolute dog and the author doesn't seem interested in speeding it up.
        # I might have to replace it with something else but in the meantime I desperately need
        # more speed than this so I'll spawn sftp manually.  If the file is over a certain
        # size.

        MAX_SIZE = 40 * 1024

        for lname, rname in files.items():
            rstat = self._stat(rname)

            if rstat and rstat.isfile:
                hash = hash_file(lname)
                if hash == rstat.hash:
                    logger.debug('Using cached file %s', rname)
                    continue

            lstat = os.stat(lname)

            if lstat.st_size <= MAX_SIZE:
                logger.debug('copying %s --> %s', lname, rname)
                sftp.put(lname, rname)
                logger.debug('copy complete')
            else:
                logger.debug('SFTP copying %s --> %s', lname, rname)
                cmd = f"scp -CBq '{lname}' '{self.host}:{rname}'"
                logger.debug('cmd: %s', cmd)
                n = os.system(cmd)
                if n:
                    raise Exception('Copy failed: {lname} -> {rname}.  error={n}')


class File:
    """
    Used to pass a local file to a remote function.

    If you pass a instance of File to a remote function, the file will be
    transferred to a temporary work directory on the remote host and the
    argument will be replaced with the remote filename.  The remote function
    receives only a string that is the fully qualified path to the file on the
    remote host.
    """
    def __init__(self, filename):
        """
        filename
          The local file to the path to send.
        """
        if not exists(filename):
            raise FileNotFoundError(filename)
        self.filename = filename

    def __repr__(self):
        return self.filename


class Template(File):
    """
    A subclass of File that first performs substitution using `str.format`.

    A local, temporary copy of the file is made after substitution and the copy
    is sent to the remote host.
    """
    def __init__(self, filename, vars):
        """
        filename
          The local file to the path to send.

        vars
          Values to substitute in the template.  This value is passed to
          `str.format` and is usually a dictionary.

          For convenience, namedtuples are passed as dictionaries in addition
          to be being passed as a tuple.  This allows fields in the tuple to be
          accessed by position ("{0}") but also by name ("{user}") which is
          considerably easier to maintain.
        """
        # I can't read from a NamedTemporaryFile after writing to it on macOS, so I'm leaving
        # it open for now.
        args = ()
        if hasattr(vars, '_asdict'):
            args = vars
            vars = vars._asdict()
        if not isinstance(vars, dict):
            raise TypeError('vars must be a dict or namedtuple')

        self.tempfile = tempfile.NamedTemporaryFile(delete=False)
        text = open(filename).read()
        text = text.format(*args, **vars)
        self.tempfile.write(text.encode('utf-8'))
        self.tempfile.close()

        File.__init__(self, self.tempfile.name)

    def __del__(self):
        os.unlink(self.tempfile.name)

    def __repr__(self):
        return 'Template(%s)' % self.filename


class FileProxy:
    """
    Represents a Python script sent to the remote server.

    Functions in the script are accessible by name: `proxy.func()`

    Only functions are supported.  Do not try to access remote variables -
    pepperssh will try to call the value as a function on the remote host.
    """
    def __init__(self, client, filename):
        mn = basename(filename)
        if mn.endswith('.py'):
            mn = mn[:-3]

        self._client = client
        self._filename = filename
        self._modulename = mn

    def __getattr__(self, name):
        proxy = FunctionProxy(self, name)
        setattr(self, name, proxy)
        return proxy


class FunctionProxy:
    """
    Represents a function in a remote script.

    Instances of these are returned for any unknown attribute accessed on the
    remote script: `x = remote.xyz()`.  Calling this object as a function sends
    an RPC request to the server and returns its response.
    """
    def __init__(self, fproxy, name):
        self.fproxy = fproxy
        self.name   = name
        self.timeout = 60.0 * 2  # two minutes (in seconds)

    def __call__(self, *args, **kwargs):
        msg = self.fproxy._client._call(
            self.fproxy._modulename,
            self.name, args, kwargs,
            timeout=self.timeout)
        return msg.get('result')



# pytest tests for the wildcard matching in archive globs.

# Use python -m pytest from the root directory.


from pepperssh.globs import translate_pattern

def _match(regexp, filenames):
    r = translate_pattern(regexp)

    filenames = [fn.strip() for fn in filenames.strip().split('\n') if fn.strip()]

    for fn in filenames:
        assert r.search(fn), f'No match: {regexp} {fn!r}'


def _dont_match(regexp, filenames):
    r = translate_pattern(regexp)

    filenames = [fn.strip() for fn in filenames.strip().split('\n') if fn.strip()]

    for fn in filenames:
        assert r.search(fn) is None, f'Invalid match: {regexp} {fn!r}'


def test_leading_double():

    _match('**/b',
           """
           b
           x/b
           x/y/z/b
           """)

    _dont_match('**/b',
                """
                ab
                x/ab
                a/b/c
                b/c
                """)

    # Since a double-star includes subdirectories, repeating it has no effect.  We allow it
    # but treat it as if it were a single: 'a/**/**/**/b' => 'a/**/b'
    _match('**/**/b',
           """
           b
           x/b
           x/y/z/b
           """)

    _dont_match('**/**/b',
                """
                ab
                x/ab
                a/b/c
                b/c
                """)


def test_filename():
    # Just a filename should match the file anywhere.
    _match('b',
           """
           b
           x/b
           x/y/z/b
           """)

    _dont_match('b',
                """
                ab
                x/ab
                b/a
                """)


def test_filename_wildcard():
    # Just a pattern with no slashes or double stars should match a filename in any
    # directory.

    _match('*.log',
           """
           .log
           test.log
           a/b/c/test.log
           """)

    _dont_match('*.log',
                """
                test.txt
                """)


def test_subdir_wildcard():
    _match('a*b/test.log',
           """
           ab/test.log
           axb/test.log
           x/y/ab/test.log
           x/axyzb/test.log
           """)

    # The star does not match '/'.  It only matches a single directory.
    _dont_match('a*b/test.log',
                """
                a/b/test.log
                ax/b/test.log
                """)


def test_trailing_star():
    # When a star is trailing, it can match any file in the previously named directory.
    _match('a/*',
           """
           a/x
           x/y/a/z
           """)

    _dont_match('a/*',
                """
                a
                x/a
                a/b/c
                """)


def test_question_star():
    "Mixing question marks and stars creates minimum length wildcards"
    # Mixing question marks and stars can be used to create a wildcard with a minimum number of
    # characters: ???* is a 3+ character wildcard.  If you mix in more stars, we ignore them
    # since they always match "zero or more".

    _match("?*.log",
           """
           test.log
           t.log
           a/b/c/t.log
           """)

    _dont_match("?*.log",
                """
                .log
                test.txt
                """)


def test_question():
    "A question mark matches a single character"
    _match("t?.log",
           """
           t1.log
           a/b/c/t1.log
           """)

    _dont_match("t?.log",
                """
                t12.log
                a/b/c/t12.log
                """)

def test_leading_slash():
    "A leading slash matches at the root"
    _match("/*.log",
           """
           .log
           test.log
           """)

    _dont_match("/*.log",
                """
                x/test.log
                x/y/z/test.log
                """)

    _match("/a/b/c/*.log",
           """
           a/b/c/.log
           a/b/c/test.log
           """)

    _dont_match("/a/b/c/*.log",
                """
                .log
                test.log
                a/b/test.log
                """)


def test_trailing_double_stars():
    """
    Trailing double-stars mean anything below matches.
    """
    _match("a/**",
           """
           a/test.log
           a/b/c/test.log
           b/c/a/test.log
           """)

    _match("/a/**",
           """
           a/test.log
           a/b/c/test.log
           """)

    _dont_match("a/**",
                """
                test.log
                x/z/y/test.log
                """)

    _dont_match("/a/**",
                """
                b/c/a/test.log
                """)

def test_directory_double_stars():
    "Double stars in middle match zero or more directories"
    _match("a/**/b",
           """
           a/b
           a/c/b
           a/c/d/b
           a/b/b
           x/y/a/b
           x/y/a/z/b
           """)

    _dont_match("a/**/b",
                """
                a/b/c
                ab
                """)




# The SSH client that sends message to the remote server.

import logging, os, re, sys, functools, socket
import paramiko
from os.path import join, abspath, dirname

from .protocol import encode, read_msg, ProtocolError

logger = logging.getLogger('pepperssh')


def _disable_logging():
    # I'm not sure how Python logging is supposed to work.  The paramiko library logs to
    # *info*.  I guess they expect that you always only log warnings in production?
    #
    # If it doesn't look like logging was specifically set for paramiko, change it to
    # warning instead of info.
    l = logging.getLogger('paramiko')
    if l.level == 0:
        l.setLevel(logging.WARNING)


class ClientConnection:
    def __init__(self, host):
        _disable_logging()

        self.host = host
        self.sshclient = None

        self.rpath = None
        # The temporary library directory on the remote server.  Our package has been copied
        # into this directory, so the directory should be added to the system path.

        self.channel = None
        # The open channel we use to send commands and read responses.

    def connect(self):
        self.sshclient = paramiko.SSHClient()
        self.sshclient._policy = paramiko.WarningPolicy()
        self.sshclient.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        cfg = self.get_host_config()
        logger.debug('self.sshclient config: %r', cfg)

        self.sshclient.connect(**cfg)

        self._copypackage()
        self._bootstrap()

    def close(self):
        if self.sshclient:
            try:
                self.sshclient.close()
            except:
                pass
        self.sshclient = None
        self.channel = None

    def _copypackage(self):
        """
        Called during connect to copy the files necessary to bootstrap our
        server on the remote box.
        """
        # Copy our package to a temporary directory to be imported.
        #
        # Since all of our files are in the same directory we'll simply enumerate them and copy
        # them.

        rin, rout, rerr = self.sshclient.exec_command("python3 -c 'import tempfile; print(tempfile.mkdtemp())'")
        rpath = rout.read().strip().decode('utf-8')
        logger.debug('remote path=%s', rpath)

        sftp = self.sshclient.open_sftp()
        sftp.chdir(rpath)

        # Make the work directory world readable for now.  If the user is changed via sudo, we
        # want it to be able to read the files.
        #
        # TODO: Redesign the security aspects.  Perhaps we should make an easy way to change
        # the ownership, either in remote_file, or on the remote side.

        sftp.chmod(rpath, 0o777)

        # Create the directory where we'll put files we copy.

        sftp.mkdir('files')

        sftp.mkdir('pepperssh')
        lpath = dirname(abspath(__file__))
        for filename in os.listdir(lpath):
            lname = join(lpath, filename)
            if not lname.endswith('.py'):
                continue
            rname = join(rpath, 'pepperssh', filename)
            logger.debug('copying %s --> %s', lname, rname)
            sftp.put(lname, rname)

        sftp.close()

        self.sshclient = self.sshclient
        self.rpath     = rpath

    def _bootstrap(self):
        """
        Called during connect to start the Python server code that will listen
        for messages.
        """
        tran = self.sshclient.get_transport()
        self.channel = tran.open_session()
        self.channel.set_combine_stderr(True)

        # WARNING: Do *not* use self.channel.get_pty().  It causes stdin to be echoed to
        # stdout, so everything we send comes right back, followed by the response.

        # Note the "-u" for unbuffered output.

        cmd = """
              python3 -uc
              'import sys;
               sys.path.insert(0, "%s");
               from pepperssh import server;
               server.main()'
              """ % self.rpath
        cmd = re.sub(r'\s*[\n]\s*', ' ', cmd).strip()
        logger.debug('cmd: %s', cmd)
        self.channel.exec_command(cmd)

    def send_command(self, msg, *, timeout):
        """
        Sends a message (a dict) and waits for the response (also a dict).
        """
        logger.debug('send: %r', msg)
        header, msgout = encode(msg)
        self.channel.sendall(header)
        self.channel.sendall(msgout)

        result = self._read_response(timeout=timeout)

        if 'stdout' in result:
            stdout = result['stdout'].strip()
            if stdout:
                print(stdout)

        if result['msgtype'] == 'error':
            errormsg = ['An error occurred on the server']
            if 'error' in result:
                errormsg[0] += ': ' + result['error']
            if 'traceback' in result:
                errormsg.append(result['traceback'])

            sys.exit('An error occurred on the server: ' + '\n'.join(errormsg))

        return result

    def _read_response(self, *, timeout):
        """
        Returns the decoded response as a dictionary.
        """
        self.channel.settimeout(timeout)
        try:
            return read_msg(functools.partial(self.channel.recv, 50 * 1024))
        except ProtocolError as ex:
            # We tried to decode the header but the data isn't one of our messages.  This
            # usually means that (1) and exception occurred before we setup comms or (2) an
            # external utility printed to stderr without it being redirected.  Either way,
            # assume it was a failure and that the buffer contains an error message / trace.

            # Something is writing one line at a time, so we if we can get more.
            self.channel.settimeout(0.5)
            buffer = ex.data
            try:
                while 1:
                    data = self.channel.recv(50 * 1024)
                    if not data:
                        break
                    buffer += data
            except socket.timeout:
                pass

            try:
                buffer = buffer.decode('utf-8')
            except UnicodeDecodeError:
                pass

            logger.error('Unexpected data on the wire: %r', buffer)
            return {
                'msgtype': 'error',
                'error': buffer
            }

        finally:
            self.channel.settimeout(None)

    def get_host_config(self):
        cfg = {}
        cfg['hostname'] = self.host

        fqn = os.path.expanduser('~/.ssh/config')
        if os.path.exists(fqn):
            f = open(fqn)
            sc = paramiko.SSHConfig()
            sc.parse(f)

            hostconfig = sc.lookup(self.host)
            if hostconfig:
                for fromkey, tokey in CONFIGMAP.items():
                    if fromkey in hostconfig:
                        cfg[tokey] = hostconfig[fromkey]

            if 'key_filename' in cfg:
                cfg['key_filename'] = cfg['key_filename'][0]

        return cfg


CONFIGMAP = {
    # Maps from lowercased keywords in an OpenSSH config file to paramiko SSHClient
    # parameter names.
    'identityfile': 'key_filename',
    'user': 'username',
    'port': 'port'
}




# Implementation of globbing for the archive function.
#
# This is almost the same as fnmatch.translate but handles '**'.  The built-in glob module
# does support it, but manually instead of a more advanced regular expression.
# For now we'll have to reimplement.

import re


NAMECHARS = r'[^/]'


def translate_pattern(pattern):
    """
    Convert a glob pattern, including support for '**', to a regular expression.
    """

    # The only double-asterisk patterns we allow are:
    #
    # - **/test.log - matches in any directory including root.  This is the same
    #   as just "test.log", so the easiest thing to do is to delete it.
    #
    # - /**/test.log - Same as above, so delete.
    #
    # - a/**/test.log - Match 'a', then any file named test.log below including
    #   "a/test.log".  A leading slash is irrelevant here.
    #
    # - a/** - match everything under a, including files in subdirectories.

    parts = []

    if pattern.startswith('/'):
        # This means it matches starting at the root directory.  Since we're
        # going to match against a string that starts under the root
        # directory, we'll replace it with "^".
        pattern = pattern[1:]
        parts.append(r'^')
    else:
        parts.append('(^|/)')

    if pattern.endswith('/'):
        # This just means the user is trying to exclude an entire directory.
        # Since we match against files, exclude everything from their down.
        pattern += '*'

    # Collapse repeated double stars which have no meaning.
    pattern = re.sub(r'\*{2}(/\*{2})+', '**', pattern)

    # Leading double starts don't many anything either, so remove those.
    #  pattern = re.sub(r'^ (\*{2} /)+', '([^/]+/)?', pattern, flags=re.VERBOSE)
    if pattern.startswith('**/'):
        parts.append('([^/]+/)*')
        pattern = pattern[3:]

    components = pattern.split('/')

    # I need to know if is the first or last component, so I'll use an index.
    for idx, comp in enumerate(components):
        isfirst = (idx == 0)
        islast  = (idx == len(components) - 1)

        if comp == '**':
            assert not isfirst
            if islast:
                # We're only going to match against files, so we can match
                # anything from here down.  But it must match *something*, so
                # use '+'.
                parts.append('.+')
            else:
                # A double-star in the middle matches *zero* or more
                # subdirectories.  The zero is what complicated this entire
                # function.  We can match nothing, but we must have a slash
                # afterwards.  Note that this will match invalid names like
                # "//a", but since are matching against real paths it is not an
                # issue.
                parts.append('([^/]+/)*')
        else:
            # This is a directory or filename component.
            parts.append(_translate_component(comp))
            if not islast:
                parts.append('/')
            else:
                parts.append('$')

    return re.compile(''.join(parts))


def _translate_component(pattern):
    """
    Translates a pattern for a single path component, which is a filename or a
    directory name.
    """
    # We need to escape everything except '?' and '*'.  I can't see an easy
    # way to do so that doesn't introduce other problems, so I'll simply
    # split on them.  We'll use capturing parens to ensure we get the text
    # between them, and the special characters themselves.
    #
    # Note that this causes empty entries at the beginning and end if they
    # start with special characters: *.log => ['', '*', '.log].  They can be
    # discarded.
    #
    # We accept multiple wildcards like '?*?'.  This example is silly, but
    # since '*' technically means "zero or more', '?*' could be useful.  We
    # simply require the wildcard match to have at least a character for
    # each '?', so "?***?***" is just 2 or more characters.

    result = []

    parts = re.split(r'([?*]+)', pattern)

    for part in parts:
        if not part:
            continue

        if '*' in part or '?' in part:

            # Definitely a wildcard, so start with the wilcard characters.  Then
            # we'll append items to determine the number of these allowed.

            # NB: At one point I had complex code to generate a more readable
            # regular expression, such as "X+" instead of "X{1,1}".  Since this
            # isn't really for human consumption I've simplified to always use
            # {min,max}.

            question_marks = part.count('?')
            minchars = str(question_marks)
            maxchars = ('' if '*' in part else minchars)

            exp = '%s{%s,%s}' % (NAMECHARS, minchars, maxchars)

            result.append(exp)
        else:
            # A text portion.  Escape it in case it has special regexp characters.
            result.append(re.escape(part))

    return ''.join(result)




import pickle, struct, logging, hashlib

MSG_PICKLED = 0x50535031        # PSP1 (pepperssh pickled v1)
FORMAT = '@LL'                  # Magic + length
HEADER_SIZE = struct.calcsize(FORMAT)

HASH_BUFFER_SIZE = 128 * 1024


logger = logging.getLogger('pepperssh')


class ProtocolError(Exception):
    """
    Raised when the data on the wire is not a pickled message.
    """
    def __init__(self, data):
        Exception.__init__(self, 'Unexpected data was received.  Perhaps stderr was not wrapped')
        self.data = data


def read_msg(readfunc):
    """
    Returns the next message read from the given function.

    Calls `readfunc` continuously (so make sure it is a blocking function) until a complete
    message has been read.
    """
    buffer = b''
    msglen = None
    while msglen is None or msglen > len(buffer):
        logger.debug('read_msg: About to read. msglen=%s bufferlen=%s', msglen, len(buffer))
        data = readfunc()
        logger.debug('read_msg: read %s bytes', len(data))
        if not data:
            raise RuntimeError('Connection closed before all data received')

        buffer += data

        if not msglen and len(buffer) >= HEADER_SIZE:
            # We have read enough for a header.  Verify the message type and extract
            # the length.
            msglen, buffer = decode_header(buffer)
            logger.debug('read_msg: decoded header.  msglen=%s', msglen)

    logger.debug('Read buffer: %r', buffer)

    msg = decode_msg(buffer)
    logger.debug('recv: %r', msg)
    return msg


def encode(msgdict):
    """
    Encodes a dictionary as a binary message.
    """
    msgbytes = pickle.dumps(msgdict)
    header = struct.pack(FORMAT, MSG_PICKLED, len(msgbytes))
    return (header, msgbytes)


def decode_header(buffer):
    if len(buffer) < HEADER_SIZE:
        raise ValueError(f'The header must be at least {HEADER_SIZE} bytes, not {len(buffer)}')
    magic, msglen = struct.unpack_from(FORMAT, buffer)
    if magic != MSG_PICKLED:
        raise ProtocolError(buffer)
    return msglen, buffer[HEADER_SIZE:]


def decode_msg(msgbytes):
    """
    Decodes a message back to a dictionary.
    """
    return pickle.loads(msgbytes)


def hash_file(filename):
    sha1 = hashlib.sha1()
    fd = open(filename, 'rb')
    while True:
        data = fd.read(HASH_BUFFER_SIZE)
        if not data:
            break
        sha1.update(data)
    return sha1.hexdigest()



#!/usr/bin/env python

import sys
from distutils.core import setup, Command

VERSION = '2.2.2'


class TestCommand(Command):
    user_options = [('pytest-args=', 'a', 'Arguments to pass to pytest')]

    def initialize_options(self):
        self.pytest_args = ''

    def finalize_options(self):
        pass

    def run(self):
        import shlex
        try:
            import pytest
        except ImportError:
            sys.exit('pytest not found (Try pip install pytest)')
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


setup(
    name='pepperssh',
    description='Simple SSH based remote execution',
    version=VERSION,
    #  package_dir={'': 'pepperssh'},
    packages=['pepperssh'],
    author='Michael Kleehammer',
    author_email='michael@kleehammer.com',
    url='https://gitlab.com/mkleehammer/pepperssh',
    cmdclass={'test': TestCommand}
)
